#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Analysis;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  /// <summary>
  /// Calculates the accumulated frequencies of variable-symbols over all trees in the population.
  /// </summary>
  [Item("VariableFrequencyAnalyser", "Calculates the accumulated frequencies of variable-symbols over all trees in the population.")]
  [StorableClass]
  public class VariableFrequencyAnalyser : SymbolicDataAnalysisAnalyzer {
    private const string VariableFrequenciesParameterName = "VariableFrequencies";
    private const string AggregateLaggedVariablesParameterName = "AggregateLaggedVariables";

    #region parameter properties
    public ILookupParameter<DataTable> VariableFrequenciesParameter {
      get { return (ILookupParameter<DataTable>)Parameters[VariableFrequenciesParameterName]; }
    }
    public IValueLookupParameter<BoolValue> AggregateLaggedVariablesParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters[AggregateLaggedVariablesParameterName]; }
    }
    #endregion
    #region properties
    public DataTable VariableFrequencies {
      get { return VariableFrequenciesParameter.ActualValue; }
      set { VariableFrequenciesParameter.ActualValue = value; }
    }
    public BoolValue AggregateLaggedVariables {
      get { return AggregateLaggedVariablesParameter.ActualValue; }
    }
    #endregion
    [StorableConstructor]
    protected VariableFrequencyAnalyser(bool deserializing) : base(deserializing) { }
    protected VariableFrequencyAnalyser(VariableFrequencyAnalyser original, Cloner cloner)
      : base(original, cloner) {
    }
    public VariableFrequencyAnalyser()
      : base() {
      Parameters.Add(new LookupParameter<DataTable>(VariableFrequenciesParameterName, "The relative variable reference frequencies aggregated over all trees in the population."));
      Parameters.Add(new ValueLookupParameter<BoolValue>(AggregateLaggedVariablesParameterName, "Switch that determines whether all references to a variable should be aggregated regardless of time-offsets. Turn off to analyze all variable references with different time offsets separately.", new BoolValue(true)));
    }

    public override IOperation Apply() {
      ItemArray<SymbolicExpressionTree> expressions = SymbolicExpressionTreeParameter.ActualValue;
      ResultCollection results = ResultCollection;

      if (VariableFrequencies == null) {
        VariableFrequencies = new DataTable("Variable frequencies", "Relative frequency of variable references aggregated over the whole population.");
        VariableFrequencies.VisualProperties.XAxisTitle = "Generation";
        VariableFrequencies.VisualProperties.YAxisTitle = "Relative Variable Frequency";
        results.Add(new Result("Variable frequencies", VariableFrequencies));
      }

      // all rows must have the same number of values so we can just take the first
      int numberOfValues = VariableFrequencies.Rows.Select(r => r.Values.Count).DefaultIfEmpty().First();

      foreach (var pair in VariableFrequencyAnalyser.CalculateVariableFrequencies(expressions, AggregateLaggedVariables.Value)) {
        if (!VariableFrequencies.Rows.ContainsKey(pair.Key)) {
          // initialize a new row for the variable and pad with zeros
          DataRow row = new DataRow(pair.Key, "", Enumerable.Repeat(0.0, numberOfValues));
          row.VisualProperties.StartIndexZero = true;
          VariableFrequencies.Rows.Add(row);
        }
        VariableFrequencies.Rows[pair.Key].Values.Add(pair.Value);
      }

      // add a zero for each data row that was not modified in the previous loop 
      foreach (var row in VariableFrequencies.Rows.Where(r => r.Values.Count != numberOfValues + 1))
        row.Values.Add(0.0);


      //foreach (var pair in VariableFrequencyAnalyser.CalculateVariableFrequencies(expressions, inputVariables)) {
      //  VariableFrequencies.Rows[pair.Key].Values.Add(pair.Value);
      //  results["Variable frequencies"].Value = VariableFrequencies;
      //}

      return base.Apply();
    }

    public static IEnumerable<KeyValuePair<string, double>> CalculateVariableFrequencies(IEnumerable<ISymbolicExpressionTree> trees, bool aggregateLaggedVariables = true) {
      Dictionary<string, double> variableFrequencies = new Dictionary<string, double>();
      int totalNumberOfSymbols = 0;

      foreach (var tree in trees) {
        var variableReferences = GetVariableReferences(tree, aggregateLaggedVariables);
        foreach (var pair in variableReferences) {
          if (variableFrequencies.ContainsKey(pair.Key)) {
          }
          totalNumberOfSymbols += pair.Value;
        }
      }

      foreach (var pair in variableFrequencies)
        yield return new KeyValuePair<string, double>(pair.Key, pair.Value / totalNumberOfSymbols);
    }

    //public static IEnumerable<KeyValuePair<string, double>> CalculateVariableFrequencies(IEnumerable<SymbolicExpressionTree> trees) {
    //  Dictionary<string, double> variableReferencesSum = new Dictionary<string, double>();
    //  Dictionary<string, double> variableFrequencies = new Dictionary<string, double>();
    //  foreach (var inputVariable in inputVariables)
    //    variableReferencesSum[inputVariable] = 0.0;
    //  foreach (var tree in trees) {
    //    var variableReferences = GetVariableReferenceCount(tree, inputVariables);
    //    foreach (var pair in variableReferences) {
    //      variableReferencesSum[pair.Key] += pair.Value;
    //    }
    //  }
    //  double totalVariableReferences = variableReferencesSum.Values.Sum();
    //  foreach (string inputVariable in inputVariables) {
    //    double relFreq = variableReferencesSum[inputVariable] / totalVariableReferences;
    //    variableFrequencies.Add(inputVariable, relFreq);
    //  }
    //  return variableFrequencies;
    //}

    private static IEnumerable<KeyValuePair<string, int>> GetVariableReferences(ISymbolicExpressionTree tree, bool aggregateLaggedVariables = true) {
      Dictionary<string, int> references = new Dictionary<string, int>();
      int currentTimeLag = 0;
      tree.Root.ForEachNodePrefix(n => {
        if (n.Symbol is LaggedVariable) {
        }
      });
      return references;
    }
  }
}
